type AsyncFunction<A extends Array<any>, PromiseReturn> = (...args: A) => Promise<PromiseReturn>;
type CoroutineGenerator<T> = Generator<any, T, any>;
type CoroutineFunction<A extends Array<any>, T> = (...args: A) => CoroutineGenerator<T>;
/**
 * A convenient utility function to simplify waiting for async. Can be used
 * in both decorator and function form. When applied to an async function, it
 * will cause tests to wait until the returned promise has resolves. When
 * applied to a generator function, it will cause tests to wait until the
 * returned iterator has run to completion, which is useful for wrapping
 * ember-concurrency task functions.
 *
 *
 * @public
 * @param promise {Function} An async function or a generator function
 * @param label {string} An optional string to identify the promise
 *
 * @example
 *
 * import Component from '@ember/component';
 * import { waitFor } from '@ember/test-waiters';
 *
 * export default Component.extend({
 *   doAsyncStuff: waitFor(async function doAsyncStuff() {
 *     await somethingAsync();
 *   }
 * });
 *
 * @example
 *
 * import Component from '@ember/component';
 * import { waitFor } from '@ember/test-waiters';
 *
 * export default class Friendz extends Component {
 *   @waitFor
 *   async doAsyncStuff() {
 *     await somethingAsync();
 *   }
 * }
 *
 */
export default function waitFor<A extends Array<any>, PromiseReturn>(fn: AsyncFunction<A, PromiseReturn>, label?: string): AsyncFunction<A, PromiseReturn>;
export default function waitFor<A extends Array<any>, T>(fn: CoroutineFunction<A, T>, label?: string): CoroutineFunction<A, T>;
export default function waitFor(target: object, _key: string, descriptor: PropertyDescriptor, label?: string): PropertyDescriptor;
export {};
//# sourceMappingURL=wait-for.d.ts.map